using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using DG.Tweening;

public class Player : MonoBehaviour {

    public enum State {
        None,
        Idle,
        Moving,
        Digging,
        Jumping
    }

    public System.Action<int> OnDiggletGained;

    [Header("References")]

    public Rigidbody2D rb;
    public PlayerSettings ps;
    public Transform visual;
    public SpriteRenderer shadow;
    public DigController digController;
    public SpawnVFX spawnVFX;
    public int playerId = 0;
    public SpriteRenderer spriteControls;
    public AudioSource moveSource;

    [Header("State")]

    public State state = State.None;
    public float height = 0;

    [Header("Stats")]

    public Vector2 lastInput;
    public Vector2 direction;
    public Vector3Int closestCell;
    private Map map;
    public Vector2 navigatePos;
    public bool navigating = false;
    public float digProgress = 0;
    public float endDigProgress = 0;
    private Hole lastHole;
    private bool controlsFadeStarted = false;

    private List<Digglet> digglets = new List<Digglet>();

    private void Awake() {
        state = State.Idle;
        map = FindObjectOfType<Map>();
    }

    // Update is called once per frame
    void FixedUpdate() {
        if (state == State.Digging || state == State.Jumping) {
            if (moveSource != null && moveSource.isPlaying) moveSource.Pause();

            return;
        }

        Vector2 input = Vector2.zero;
        if (navigating) {
            Vector2 playerPos = rb.position;
            input = (navigatePos - playerPos).normalized;
        } else {
            float h = Input.GetAxis(playerId == 0 ? "Horizontal" : "Horizontal2");
            float v = Input.GetAxis(playerId == 0 ? "Vertical" : "Vertical2");

            input = new Vector2(h, v);
            if (input.magnitude > 1) {
                input /= input.magnitude;
            }
        }
        Vector2 force = input * ps.acceleration * Time.fixedDeltaTime;

        lastInput = input;

        state = input.magnitude > 0.1f ? State.Moving : State.Idle;
        if (state == State.Moving) {
            direction = lastInput.normalized;
            FadeOutControls();
            if (moveSource != null && !moveSource.isPlaying) moveSource.Play();
        } else {
            if (moveSource != null && moveSource.isPlaying) moveSource.Pause();
        }

        rb.MovePosition(rb.position + force);

        if (navigating) {
            if (Vector2.Distance(rb.position, navigatePos) < ps.acceleration * Time.fixedDeltaTime) {
                navigating = false;
                OnNavigated();
            }
        }
    }

    private void FadeOutControls() {
        if (!controlsFadeStarted) {
            controlsFadeStarted = true;
            spriteControls.transform.DOScale(0f, 0.3f).SetEase(Ease.InBack).SetDelay(1f);
            DOVirtual.DelayedCall(2f, GameController.instance.HideLogo);
        }
    }

    public Transform GetTargetForDigglet() {
        if (digglets.Count == 0)
            return transform;
        else
            return digglets[digglets.Count - 1].transform;
    }

    public void AddDigglet(Digglet digglet) {
        digglets.Add(digglet);
        OnDiggletGained?.Invoke(digglets.Count - 1);
    }

    private void Update() {
        if (Input.GetButtonDown(playerId == 0 ? "Jump" : "Jump2")) {
            OnAction();
        }

        if (Input.GetKeyDown(KeyCode.C)) {
            Digglet digglet = Instantiate(GameController.instance.eggs[0].diggletPrefab, transform.position, Quaternion.identity);
            digglet.Setup(this);
        }

        if (state == State.Digging) {
            //digProgress += Time.deltaTime / 2;
            //if (digProgress >= 1) {
            //    OnDigFinished();
            //}

            height = digProgress * -ps.digDepth;
        }

        UpdateVisualPos();
    }

    private void OnAction() {
        if (state == State.Jumping) return;

        if (state == State.Digging) {
            digController.OnAction();
        } else if (!navigating) {
            NavigateToClosestCell();
        }
    }

    private void UpdateVisualPos() {
        visual.localPosition = visual.localPosition.SetY(height);
        shadow.transform.localScale = Vector3.one * (height < 0 ? 0 : height.Map(0, ps.jumpHeight, 1f, 0.2f));
        shadow.color = shadow.color.SetA(height.Map(0, ps.jumpHeight, 0.5f, 0f));
    }

    private void NavigateToClosestCell() {
        closestCell = map.GetClosestCell(rb.position);
        Cell cell = map.GetCell(closestCell);
        if (cell.dug) return; // already dug

        Vector3 pos = map.GetClosestCellCenter(rb.position);
        navigatePos = pos;
        navigating = true;
    }

    private void OnNavigated() {
        Cell cell = map.GetCellFromWorld(rb.position);
        if (!cell.dug) {
            StartDigging();
        }
    }

    private void StartDigging() {
        state = State.Digging;
        digProgress = 0;
        endDigProgress = 0;
        lastHole = Instantiate(GameController.instance.hole, transform.position, Quaternion.identity);

        if (ps.instaDig) {
            AddDigProgress(1);
        } else {
            digController.StartDigging();
        }
    }

    private void OnDigFinished() {
        Cell cell = map.GetCellFromWorld(rb.position);
        cell.dug = true;
        map.OnHoleDug?.Invoke(map.GetClosestCell(rb.position));

        if (cell.egg) {
            OnEggFound();
        } else {
            OnArrowFound();
        }

        state = State.Jumping;


        Sequence seq = DOTween.Sequence()
        .Append(DOVirtual.Float(height, ps.jumpHeight, ps.jumpDurationUp, (f) => { height = f; }).SetEase(Ease.OutQuad))
        .Append(DOVirtual.Float(ps.jumpHeight, 0, ps.jumpDurationDown, (f) => { height = f; }).SetEase(Ease.InQuad))
        .OnComplete(() => state = State.Idle);

        Vector3 rng = Random.insideUnitCircle;
        rng /= rng.magnitude;
        rng *= 1.5f;
        transform.DOMove(transform.position + rng, ps.jumpDurationUp + ps.jumpDurationDown).SetEase(Ease.Linear);

        DOVirtual.DelayedCall(0.2f, () => {
            lastHole.DisableMask();
        });
    }

    private void OnEggFound() {
        Egg egg = Instantiate(GameController.instance.eggs.Random(), visual.position, Quaternion.identity);
        egg.Setup(this);
    }

    private void OnArrowFound() {
        Arrow arrow = Instantiate(GameController.instance.arrow, transform.position, Quaternion.identity);
        arrow.Setup();
    }

    public void AddDigProgress(float progress) {
        endDigProgress = digProgress + progress;
        DOVirtual.Float(digProgress, endDigProgress, 0.3f, (f) => digProgress = f).SetEase(Ease.OutQuad).SetDelay(0.2f);

        lastHole.MakeBig();

        if (endDigProgress >= 1) {
            DOVirtual.DelayedCall(1f, OnDigFinished);
        }
    }
}
